from urllib.parse import urljoin

import requests
from requests import Response, HTTPError

from biolib.typing_utils import Dict, Optional, Union
from biolib.biolib_api_client import BiolibApiClient as DeprecatedApiClient


class ApiClient:

    def get(
            self,
            url: str,
            params: Optional[Dict[str, Union[str, int]]] = None,
            headers: Optional[Dict[str, str]] = None,
    ) -> Response:
        response = requests.get(
            headers=self._get_headers(headers),
            params=params,
            timeout=10,
            url=self._get_absolute_url(url),
        )
        ApiClient.raise_for_status(response)
        return response

    def post(self, path: str, data: Dict) -> Response:
        response = requests.post(
            headers=self._get_headers(),
            json=data,
            timeout=10,
            url=self._get_absolute_url(path),
        )
        ApiClient.raise_for_status(response)
        return response

    @staticmethod
    def raise_for_status(response):
        # Logic taken from `requests.Response.raise_for_status()`
        http_error_msg = ''
        reason = response.text
        if 400 <= response.status_code < 500:
            http_error_msg = u'%s Client Error: %s for url: %s' % (response.status_code, reason, response.url)

        elif 500 <= response.status_code < 600:
            http_error_msg = u'%s Server Error: %s for url: %s' % (response.status_code, reason, response.url)

        if http_error_msg:
            raise HTTPError(http_error_msg, response=response)

    @staticmethod
    def _get_headers(opt_headers: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        headers: Dict[str, str] = opt_headers or {}

        deprecated_api_client = DeprecatedApiClient.get()

        if deprecated_api_client.is_signed_in:
            deprecated_api_client.refresh_access_token()

        # Adding access_token outside is_signed_in check as job_worker.py currently sets access_token
        # without setting refresh_token
        access_token = deprecated_api_client.access_token
        if access_token:
            headers['Authorization'] = f'Bearer {access_token}'

        return headers

    @staticmethod
    def _get_absolute_url(path: str) -> str:
        deprecated_api_client = DeprecatedApiClient.get()
        base_api_url = urljoin(deprecated_api_client.base_url, '/api/')
        return urljoin(base_api_url, path.strip('/') + '/')



from pathlib import Path
from fnmatch import fnmatch
import time
import requests

from biolib import api
from biolib.biolib_binary_format import UnencryptedModuleOutput
from biolib.biolib_binary_format.utils import RemoteIndexableBuffer, LazyLoadedFile
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger_no_user_data, logger
from biolib.typing_utils import Optional, Dict, List, cast, Union, Callable


class JobResultError(BioLibError):
    pass


class JobResultNotFound(JobResultError):
    pass


class JobResultPermissionError(JobResultError):
    pass


PathFilter = Union[str, Callable[[LazyLoadedFile], bool]]


class JobResult:

    def __init__(
            self,
            job_uuid: str,
            job_auth_token: Optional[str],
            module_output: Optional[UnencryptedModuleOutput] = None,
    ):
        self._job_uuid: str = job_uuid
        self._job_auth_token: Optional[str] = job_auth_token

        self._module_output: Optional[UnencryptedModuleOutput] = module_output

    def get_stdout(self) -> bytes:
        return self._get_module_output().get_stdout()

    def get_stderr(self) -> bytes:
        return self._get_module_output().get_stderr()

    def get_exit_code(self) -> int:
        return self._get_module_output().get_exit_code()

    def save_files(self, output_dir: str, path_filter: Optional[PathFilter] = None) -> None:
        module_output = self._get_module_output()
        output_files = module_output.get_files()

        if path_filter:
            output_files = self._get_filtered_files(output_files, path_filter)

        logger.info(f'Saving {len(output_files)} files to {output_dir}...')

        for file in output_files:
            # Remove leading slash of file_path
            destination_file_path = Path(output_dir) / Path(file.path.lstrip('/'))
            if destination_file_path.exists():
                destination_file_path.rename(f'{destination_file_path}.biolib-renamed.{time.strftime("%Y%m%d%H%M%S")}')

            dir_path = destination_file_path.parent
            if dir_path:
                dir_path.mkdir(parents=True, exist_ok=True)

            with open(destination_file_path, mode='wb') as file_handler:
                for chunk_number, chunk in enumerate(file.get_data_as_iterable()):
                    logger_no_user_data.debug(f'Processing chunk {chunk_number}...')
                    file_handler.write(chunk)

            logger.info(f'  - {destination_file_path}')

    def get_output_file(self, filename) -> LazyLoadedFile:
        files = self._get_module_output().get_files()
        filtered_files = self._get_filtered_files(files, path_filter=filename)
        if not filtered_files:
            raise BioLibError(f"File {filename} not found in results.")

        if len(filtered_files) != 1:
            raise BioLibError(f"Found multiple results for filename {filename}.")

        return filtered_files[0]

    def list_output_files(self, path_filter: Optional[PathFilter] = None) -> List[LazyLoadedFile]:
        files = self._get_module_output().get_files()
        if not path_filter:
            return files

        return self._get_filtered_files(files, path_filter)

    @staticmethod
    def _get_filtered_files(files: List[LazyLoadedFile], path_filter: PathFilter) -> List[LazyLoadedFile]:
        if not (isinstance(path_filter, str) or callable(path_filter)):
            raise Exception('Expected path_filter to be a string or a function')

        if callable(path_filter):
            return list(filter(path_filter, files))

        glob_filter = cast(str, path_filter)

        def _filter_function(file: LazyLoadedFile) -> bool:
            return fnmatch(file.path, glob_filter)

        return list(filter(_filter_function, files))

    def _get_presigned_download_url(self) -> str:
        try:
            response_dict: Dict[str, str] = api.client.get(
                url=f'/jobs/{self._job_uuid}/storage/results/download/',
                headers={'Job-Auth-Token': self._job_auth_token} if self._job_auth_token else None,
            ).json()
            return response_dict['presigned_download_url']
        except requests.exceptions.HTTPError as error:
            status_code = error.response.status_code

            if status_code == 401:
                raise JobResultPermissionError('You must be signed in to get result of the job') from None
            elif status_code == 403:
                raise JobResultPermissionError(
                    'Cannot get result of job. Maybe the job was created without being signed in?'
                ) from None
            elif status_code == 404:
                raise JobResultNotFound('Job result not found') from None
            else:
                raise JobResultError('Failed to get result of job') from error

        except Exception as error:
            raise JobResultError('Failed to get result of job') from error

    def _get_module_output(self) -> UnencryptedModuleOutput:
        if self._module_output is None:
            buffer = RemoteIndexableBuffer(url=self._get_presigned_download_url())
            self._module_output = UnencryptedModuleOutput(buffer)

        return self._module_output



import os
from biolib.compute_node.job_worker.job_storage import JobStorage
from biolib.experiments.experiment import Experiment
from biolib.jobs import Job
from biolib.typing_utils import List, Optional, cast
from biolib.app.utils import run_job
from biolib.biolib_api_client import JobState, BiolibApiClient, CreatedJobDict
from biolib.jobs.types import JobDict
from biolib.biolib_api_client.app_types import App, AppVersion
from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
from biolib.biolib_api_client.biolib_app_api import BiolibAppApi
from biolib.biolib_binary_format import ModuleInput
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger


class BioLibApp:

    def __init__(self, uri: str):
        app_response = BiolibAppApi.get_by_uri(uri)
        self._app: App = app_response['app']
        self._app_uri = app_response['app_uri']
        self._app_version: AppVersion = app_response['app_version']

        logger.info(f'Loaded project {self._app_uri}')

    def __str__(self) -> str:
        return self._app_uri

    @property
    def uuid(self) -> str:
        return self._app['public_id']

    @property
    def version(self) -> AppVersion:
        return self._app_version

    def cli(
            self,
            args=None,
            stdin=None,
            files=None,
            override_command=False,
            machine='',
            blocking: bool = True,
            experiment_id: Optional[str] = None,
            result_prefix: Optional[str] = None,
    ):

        if not experiment_id:
            experiment = Experiment.get_experiment_in_context()
            experiment_id = experiment.uuid if experiment else None

        if not blocking:
            return self._cli_non_blocking(
                args,
                stdin,
                files,
                override_command,
                machine,
                experiment_id,
                result_prefix,
            )

        module_input_serialized = self._get_serialized_module_input(args, stdin, files)

        job: CreatedJobDict = BiolibJobApi.create(
            app_version_id=self._app_version['public_id'],
            override_command=override_command,
            experiment_uuid=experiment_id,
            machine=machine if machine != 'local' else '',
        )

        BiolibJobApi.update_state(job['public_id'], JobState.IN_PROGRESS.value)

        try:
            module_output = run_job(
                job,
                module_input_serialized,
                force_local=(machine == 'local'),
                result_name_prefix=result_prefix,
            )
            try:
                BiolibJobApi.update_state(job_id=job['public_id'], state=JobState.COMPLETED.value)
            except Exception as error:  # pylint: disable=broad-except
                logger.debug(f'Could not update job state to completed:\n{error}')

            # TODO: Create new JobDict that has shared relevant fields
            job_dict = cast(JobDict, job)
            job_object = Job(job_dict=job_dict)
            job_object._set_result_module_output(module_output)  # pylint: disable=protected-access
            return job_object

        except BioLibError as exception:
            logger.error(f'Compute failed with: {exception.message}')
            try:
                BiolibApiClient.refresh_auth_token()
                BiolibJobApi.update_state(job_id=job['public_id'], state=JobState.FAILED.value)
            except Exception as error:  # pylint: disable=broad-except
                logger.debug(f'Could not update job state to failed:\n{error}')

            raise exception

    def exec(self, args=None, stdin=None, files=None, machine=''):
        return self.cli(args, stdin, files, override_command=True, machine=machine)

    def __call__(self, *args, **kwargs):
        if not args and not kwargs:
            self.cli()

        else:
            raise BioLibError('''
Calling an app directly with app() is currently being reworked.
To use the previous functionality, please call app.cli() instead. 
Example: "app.cli('--help')"
''')

    @staticmethod
    def _get_serialized_module_input(args=None, stdin=None, files=None) -> bytes:
        if args is None:
            args = []

        if stdin is None:
            stdin = b''

        if isinstance(args, str):
            args = list(filter(lambda p: p != '', args.split(' ')))

        if not isinstance(args, list):
            raise Exception('The given input arguments must be list or str')

        if isinstance(stdin, str):
            stdin = stdin.encode('utf-8')

        if files is None:
            files = []
            for idx, arg in enumerate(args):
                if os.path.isfile(arg):
                    files.append(arg)
                    args[idx] = arg.split('/')[-1]

        cwd = os.getcwd()
        files_dict = {}
        if isinstance(files, list):
            for file in files:
                path = file
                if not file.startswith('/'):
                    # make path absolute
                    path = cwd + '/' + file

                arg_split = path.split('/')
                file = open(path, 'rb')
                path = '/' + arg_split[-1]

                files_dict[path] = file.read()
                file.close()
        elif isinstance(files, dict):
            files_dict = files
        else:
            raise Exception('The given files input must be list or dict or None')

        module_input_serialized: bytes = ModuleInput().serialize(
            stdin=stdin,
            arguments=args,
            files=files_dict,
        )
        return module_input_serialized

    def _cli_non_blocking(
            self,
            args: Optional[List[str]] = None,
            stdin: Optional[bytes] = None,
            files: Optional[List[str]] = None,
            override_command: bool = False,
            machine: Optional[str] = None,
            experiment_id: Optional[str] = None,
            result_prefix: Optional[str] = None,
    ) -> Job:
        module_input_serialized = self._get_serialized_module_input(args, stdin, files)

        job_dict: CreatedJobDict = BiolibJobApi.create(
            app_version_id=self._app_version['public_id'],
            override_command=override_command,
            machine=machine,
            experiment_uuid=experiment_id
        )
        JobStorage.upload_module_input(job=job_dict, module_input_serialized=module_input_serialized)
        cloud_job = BiolibJobApi.create_cloud_job(job_id=job_dict['public_id'], result_name_prefix=result_prefix)
        logger.debug(f"Cloud: Job created with id {cloud_job['public_id']}")
        return Job(cast(JobDict, job_dict))



import logging
from typing import Optional

import click

from biolib.biolib_logging import logger, logger_no_user_data
from biolib.biolib_push import push_application


@click.command(help='Push an application to BioLib')
@click.argument('uri')
@click.option('--path', default='.', required=False)
@click.option('--copy-images-from-version', required=False)
@click.option('--dev', is_flag=True, default=False, required=False)
def push(uri, path: str, copy_images_from_version: Optional[str], dev: bool) -> None:
    logger.configure(default_log_level=logging.INFO)
    logger_no_user_data.configure(default_log_level=logging.INFO)
    push_application(
        app_path=path,
        app_uri=uri,
        app_version_to_copy_images_from=copy_images_from_version,
        is_dev_version=dev,
    )



# type: ignore
# TODO: Remove this file and function
import sys
import click
from docker.errors import ImageNotFound  # type: ignore

from biolib import utils, cli_utils
from biolib.app.utils import run_job
from biolib.biolib_binary_format import ModuleInput
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_logging import logger, logger_no_user_data
from biolib.biolib_api_client import CreatedJobDict
from biolib.cli_utils import BiolibValidationError
from biolib.typing_utils import Tuple
from biolib.validators.validate_zip_file import validate_zip_file  # type: ignore


@click.command(
    context_settings=dict(ignore_unknown_options=True, allow_interspersed_args=False),
    help='Run an application from a local directory',
    hidden=True,
)
@click.argument('path', required=True)
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def run_dev(path: str, args: Tuple[str]) -> None:
    app_path = cli_utils.validate_and_get_app_path(path)
    logger.info(f'Running BioLib application in local directory {app_path}...')
    tmp_directory, source_files_zip_path = cli_utils.create_temporary_directory_with_source_files_zip(app_path)

    try:
        validate_zip_file(source_files_zip_path, app_path)
        yaml_data = cli_utils.get_yaml_data(app_path)
        cli_utils.validate_yaml_config(yaml_data, source_files_zip_path)
        app_version_mocked = cli_utils.get_mocked_app_version_from_yaml(yaml_data, source_files_zip_path)

        for module in app_version_mocked['modules']:
            try:
                BiolibDockerClient.get_docker_client().images.get(module['image_uri'])
            except ImageNotFound:
                raise Exception(
                    f"Could not find local docker image {module['image_uri']} specified in .biolib/config.yml"
                ) from None

        args = cli_utils.validate_and_get_args(list(args))
        files_dict, args = cli_utils.get_files_dict_and_file_args(files=None, args=args)

        def _get_stdin():
            stdin = None
            if not sys.stdin.isatty() and not utils.IS_DEV:
                stdin = sys.stdin.read()
            return stdin

        module_input_serialized = ModuleInput().serialize(
            stdin=_get_stdin() or b'',
            files=files_dict,
            arguments=args,
        )

        module_output = run_job(
            module_input_serialized=module_input_serialized,
            job=CreatedJobDict(
                app_version=app_version_mocked,
                arguments_override_command=False,
                auth_token='',
                caller_job=None,
                created_at='',
                federated_job_uuid=None,
                public_id=utils.RUN_DEV_JOB_ID,
                remote_hosts_with_warning=[],
                state='in_progress',
                user_id=None,
            ),
            force_local=True,
        )

        cli_utils.write_output_files(module_output['files'])  # type: ignore

        print(cli_utils.get_pretty_print_module_output_string(
            exitcode=module_output['exit_code'],  # type: ignore
            stderr=module_output['stderr'],  # type: ignore
            stdout=module_output['stdout'],  # type: ignore
        ))

    except Exception as exception:
        # Exit on BiolibValidationError as we have already printed the validation errors
        if isinstance(exception, BiolibValidationError):
            logger_no_user_data.error('Validation check failed for config file at .biolib/config.yml')
            exit(1)
        raise exception

    finally:
        try:
            tmp_directory.cleanup()
        except IOError:
            logger_no_user_data.error('Failed to clean up temporary source files zip directory')



from biolib.compute_node.job_worker.executors.docker_executor import DockerExecutor
from biolib.compute_node.job_worker.executors.remote import RemoteExecutor

from biolib.compute_node.job_worker.executors.types import *



import logging
import os
import sys
import random
import subprocess
import threading
import time
import pathlib
from io import BytesIO
from typing import BinaryIO

import requests.exceptions

from biolib import utils
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
from biolib.biolib_binary_format import UnencryptedModuleOutput, InMemoryIndexableBuffer
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger
from biolib.compute_node.job_worker.executors import RemoteExecutor, RemoteExecuteOptions
from biolib.utils import stream_process_output
from biolib.typing_utils import Optional


def run_job(
        job,
        module_input_serialized: bytes,
        force_local: bool = False,
        result_name_prefix: Optional[str] = None,
) -> UnencryptedModuleOutput:
    job_id = job['public_id']
    logger.info(f'Job "{job_id}" is starting...')

    if not force_local:
        return RemoteExecutor.execute_job(
            RemoteExecuteOptions(
                biolib_base_url=utils.BIOLIB_BASE_URL,
                job=job,
                result_name_prefix=result_name_prefix,
                root_job_id=job_id,
            ),
            module_input_serialized,
        )

    if not job['app_version'].get('modules'):
        BioLibError('Unable to run the application locally')

    # Run locally
    host = '127.0.0.1'
    port = str(random.choice(range(5000, 65000)))
    node_url = f'http://{host}:{port}'
    logger.debug(f'Starting local compute node at {node_url}')
    start_cli_path = pathlib.Path(__file__).parent.parent.resolve() / 'start_cli.py'
    python_executable_path = sys.executable
    output_destination: Optional[int]
    if utils.IS_RUNNING_IN_NOTEBOOK:
        utils.STREAM_STDOUT = True
        output_destination = subprocess.PIPE
    elif utils.STREAM_STDOUT:
        # If not running in notebook but streaming stdout is enabled (probably running as CLI)
        # set output destination to None to have it write to current process output (this does not work in notebook)
        output_destination = None
    else:
        output_destination = subprocess.DEVNULL

    compute_node_process = subprocess.Popen(
        args=[python_executable_path, start_cli_path, 'start', '--host', host, '--port', port],
        env=dict(
            os.environ,
            BIOLIB_LOG=logging.getLevelName(logger.level),
            BIOLIB_BASE_URL=BiolibApiClient.get().base_url
        ),
        stdout=output_destination,
        stderr=output_destination,
    )

    if utils.IS_RUNNING_IN_NOTEBOOK:
        logger.debug('Running in Notebook, so starting output streaming thread...')
        threading.Thread(target=stream_process_output, args=(compute_node_process,)).start()

    try:
        for retry in range(20):
            time.sleep(1)
            try:
                BiolibJobApi.save_compute_node_job(
                    job=job,
                    module_name='main',
                    access_token=BiolibApiClient.get().access_token,
                    node_url=node_url
                )
                break

            except requests.exceptions.ConnectionError:
                if retry == 19:
                    raise BioLibError('Could not connect to local compute node') from None
                logger.debug('Could not connect to local compute node retrying...')

        BiolibJobApi.start_local_job(job_id, module_input_serialized, node_url)
        BiolibJobApi.await_local_compute_node_status(
            retry_interval_seconds=1.5,
            retry_limit_minutes=43800,  # Let users run an app locally for a month (43800 minutes)
            status_to_await='Result Ready',
            compute_type='Compute Node',
            node_url=node_url,
            job=job,
        )

        result = BiolibJobApi.get_cloud_result(job_id, node_url)
        unencrypted_module_output: BinaryIO = BytesIO()
        legacy_module_output: BinaryIO = BytesIO(result)
        UnencryptedModuleOutput.write_to_file_from_legacy_module_output_file(
            input_file=legacy_module_output,
            output_file=unencrypted_module_output
        )
        unencrypted_module_output.seek(0)
        return UnencryptedModuleOutput(InMemoryIndexableBuffer(unencrypted_module_output.read()))

    finally:
        compute_node_process.terminate()



import io
import tarfile
import subprocess
import time

from docker.models.containers import Container  # type: ignore
from docker.errors import ImageNotFound  # type: ignore
from docker.models.images import Image  # type: ignore
from docker.models.networks import Network  # type: ignore

from biolib import utils
from biolib.biolib_errors import BioLibError
from biolib.compute_node.cloud_utils import CloudUtils
from biolib.typing_utils import Optional, List
from biolib.biolib_api_client import RemoteHost
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_logging import logger, logger_no_user_data
from biolib.biolib_api_client import BiolibApiClient


# Prepare for remote hosts with specified port
class RemoteHostExtended(RemoteHost):
    ports: List[int]


class RemoteHostProxy:
    _DOCKER_IMAGE_URI = 'nginx:1.21.1-alpine'
    _TRAFFIC_FORWARDER_PORT_OFFSET = 10000  # Port offset relative to port of a VSOCK proxy

    def __init__(
            self,
            remote_host: RemoteHost,
            public_network: Network,
            internal_network: Optional[Network],
            job_id: str,
            ports: List[int]
    ):
        self.is_app_caller_proxy = remote_host['hostname'] == 'AppCallerProxy'

        # Default to port 443 for now until backend serves remote_hosts with port specified
        self._remote_host: RemoteHostExtended = RemoteHostExtended(
            hostname=remote_host['hostname'],
            ports=ports
        )
        self._public_network: Network = public_network
        self._internal_network: Optional[Network] = internal_network

        if not job_id:
            raise Exception('RemoteHostProxy missing argument "job_id"')

        self._name = f"biolib-remote-host-proxy-{job_id}-{self.hostname}"
        self._job_uuid = job_id
        self._container: Optional[Container] = None
        self._enclave_traffic_forwarder_processes: List[subprocess.Popen] = []

    @property
    def hostname(self) -> str:
        return self._remote_host['hostname']

    def get_ip_address_on_network(self, network: Network) -> str:
        if not self._container:
            raise Exception('RemoteHostProxy not yet started')

        container_networks = self._container.attrs['NetworkSettings']['Networks']
        if network.name in container_networks:
            ip_address: str = container_networks[network.name]['IPAddress']
            return ip_address

        raise Exception(f'RemoteHostProxy not connected to network {network.name}')

    def start(self) -> None:
        # TODO: Implement nice error handling in this method

        upstream_server_name = self._remote_host['hostname']
        upstream_server_ports = self._remote_host['ports']

        docker = BiolibDockerClient.get_docker_client()

        for index in range(3):
            logger_no_user_data.debug(f'Attempt {index} at creating RemoteHostProxy container "{self._name}"...')
            try:
                self._container = docker.containers.create(
                    detach=True,
                    image=self._get_nginx_docker_image(),
                    name=self._name,
                    network=self._public_network.name,
                )
                break
            except Exception as error:  # pylint: disable=broad-except
                logger_no_user_data.exception(f'Failed to create container "{self._name}" hit error: {error}')

            logger_no_user_data.debug('Sleeping before re-trying container creation...')
            time.sleep(3)

        if not self._container:
            raise BioLibError(f'Exceeded re-try limit for creating container {self._name}')

        self._write_nginx_config_to_container(
            upstream_server_name,
            upstream_server_ports,
        )

        if self._internal_network:
            self._internal_network.connect(self._container.id)

        self._container.start()

        proxy_is_ready = False
        for retry_count in range(1, 5):
            time.sleep(0.5 * retry_count)
            # Use the container logs as a health check.
            # By using logs instead of a http endpoint on the NGINX we avoid publishing a port of container to the host
            container_logs = self._container.logs()
            if b'ready for start up\n' in container_logs or b'start worker process ' in container_logs:
                proxy_is_ready = True
                break

        if not proxy_is_ready:
            self.terminate()
            raise Exception('RemoteHostProxy did not start properly')

        self._container.reload()

    def terminate(self):
        # TODO: Implement nice error handling in this method

        if self._container:
            self._container.remove(force=True)

        for process in self._enclave_traffic_forwarder_processes:
            process.terminate()

    def _get_nginx_docker_image(self) -> Image:
        docker = BiolibDockerClient.get_docker_client()
        try:
            return docker.images.get(self._DOCKER_IMAGE_URI)
        except ImageNotFound:
            logger.debug('Pulling remote host docker image...')
            return docker.images.pull(self._DOCKER_IMAGE_URI)

    def _write_nginx_config_to_container(self, upstream_server_name: str, upstream_server_ports: List[int]) -> None:
        if not self._container:
            raise Exception('RemoteHostProxy container not defined when attempting to write NGINX config')

        docker = BiolibDockerClient.get_docker_client()
        base_url = BiolibApiClient.get().base_url
        if self.is_app_caller_proxy:
            logger_no_user_data.debug(f'Job "{self._job_uuid}" writing config for and starting App Caller Proxy...')
            if utils.BIOLIB_CLOUD_BASE_URL:
                cloud_base_url = utils.BIOLIB_CLOUD_BASE_URL
            else:
                if base_url in ('https://biolib.com', 'https://staging-elb.biolib.com'):
                    cloud_base_url = 'https://biolibcloud.com'
                else:
                    raise BioLibError('Calling apps inside apps is not supported in local compute environment')

            if utils.IS_RUNNING_IN_CLOUD:
                config = CloudUtils.get_webserver_config()
                s3_results_bucket_name = config['s3_general_storage_bucket_name']  # pylint: disable=unsubscriptable-object
                s3_results_base_url = f'https://{s3_results_bucket_name}.s3.amazonaws.com'
            else:
                if base_url in ('https://biolib.com', 'https://staging-elb.biolib.com'):
                    s3_results_base_url = 'https://biolib-cloud-api.s3.amazonaws.com'
                else:
                    raise BioLibError("Calling apps inside apps locally is only supported on biolib.com")

            # TODO: Get access_token from new API class instead
            access_token = BiolibApiClient.get().access_token
            bearer_token = f'Bearer {access_token}' if access_token else ''

            nginx_config = f'''
events {{
  worker_connections  1024;
}}

http {{
    map $request_method $bearer_token_on_post {{
        POST       "{bearer_token}";
        default    "";
    }}

    map $request_method $bearer_token_on_get {{
        GET        "{bearer_token}";
        default    "";
    }}

    map $request_method $bearer_token_on_patch {{
        PATCH      "{bearer_token}";
        default    "";
    }}

    server {{
        listen       80;
        resolver 127.0.0.11 valid=30s;

        location ~* "^/api/jobs/cloud/(?<job_id>[a-z0-9-]{{36}})/status/$" {{
            proxy_pass               {base_url}/api/jobs/cloud/$job_id/status/;
            proxy_set_header         authorization $bearer_token_on_get;
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/cloud/$" {{
            # Note: Using $1 here as URI part from regex must be used for proxy_pass
            proxy_pass               {base_url}/api/jobs/cloud/$1;
            proxy_set_header         authorization $bearer_token_on_post;
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/storage/input/start_upload/$" {{
            proxy_pass               {base_url}/api/jobs/$job_id/storage/input/start_upload/;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/storage/input/presigned_upload_url/$" {{
            proxy_pass               {base_url}/api/jobs/$job_id/storage/input/presigned_upload_url/$is_args$args;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/storage/input/complete_upload/$" {{
            proxy_pass               {base_url}/api/jobs/$job_id/storage/input/complete_upload/;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/(?<job_id>[a-z0-9-]{{36}})/$" {{
            proxy_pass               {base_url}/api/jobs/$job_id/;
            proxy_set_header         authorization $bearer_token_on_patch;
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location ~* "^/api/jobs/$" {{
            # Note: Using $1 here as URI part from regex must be used for proxy_pass
            proxy_pass               {base_url}/api/jobs/$1;
            proxy_set_header         authorization $bearer_token_on_post;
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /api/ {{
            proxy_pass               {base_url}/api/;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /cloud-proxy/ {{
            proxy_pass               {cloud_base_url}/cloud-proxy/;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /job-storage/ {{
            proxy_pass               {s3_results_base_url}/job-storage/;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}

        location /proxy/storage/job-storage/ {{
            proxy_pass               {cloud_base_url}/proxy/storage/job-storage/;
            proxy_set_header         authorization "";
            proxy_set_header         cookie "";
            proxy_ssl_server_name    on;
        }}
    }}
}}
'''
        else:
            nginx_config = '''
events {}
error_log /dev/stdout info;
stream {
    resolver 127.0.0.11 valid=30s;'''
            for idx, upstream_server_port in enumerate(upstream_server_ports):
                nginx_config += f'''
    map "" $upstream_{idx} {{
        default {upstream_server_name}:{upstream_server_port};
    }}

    server {{
        listen          {self._remote_host['ports'][idx]};
        proxy_pass      $upstream_{idx};
    }}

    server {{
        listen          {self._remote_host['ports'][idx]} udp;
        proxy_pass      $upstream_{idx};
    }}'''

            nginx_config += '''
}
'''

        nginx_config_bytes = nginx_config.encode()
        tarfile_in_memory = io.BytesIO()
        with tarfile.open(fileobj=tarfile_in_memory, mode='w:gz') as tar:
            info = tarfile.TarInfo('/nginx.conf')
            info.size = len(nginx_config_bytes)
            tar.addfile(info, io.BytesIO(nginx_config_bytes))

        tarfile_bytes = tarfile_in_memory.getvalue()
        tarfile_in_memory.close()
        docker.api.put_archive(self._container.id, '/etc/nginx', tarfile_bytes)



from .module_input import *
from .module_output import *
from .rsa_encrypted_aes_package import *
from .system_status_update import *
from .saved_job import *
from .system_exception import *
from .base_bbf_package import *
from .unencrypted_module_output import *
from .encrypted_module_output import *
